//-----------------------------------------------------------------------------
//   qualrout.c
//
//   Project:  EPA SWMM5
//   Version:  5.0
//   Date:     6/19/07   (Build 5.0.010)
//             1/21/09   (Build 5.0.014)
//   Author:   L. Rossman
//
//   Water quality routing functions.
//
//   ===============================================================
//   NOTE: This module was completely re-written for release 5.0.014
//   ===============================================================
//
//-----------------------------------------------------------------------------
#define _CRT_SECURE_NO_DEPRECATE

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "headers.h"

//-----------------------------------------------------------------------------
//  External functions (declared in funcs.h)
//-----------------------------------------------------------------------------
//  qualrout_execute         (called by routing_execute)
//  qualrout_getCstrQual     (called by getPondedQual in subcatch.c) 

//-----------------------------------------------------------------------------
//  Function declarations
//-----------------------------------------------------------------------------
static void  findLinkMassFlow(int i);
static void  findNodeQual(int j);
static void  findLinkQual(int i, double tStep);
static void  findStorageQual(int j, double tStep);
static void  updateHRT(int j, double v, double q, double tStep);


//=============================================================================

void qualrout_execute(double tStep)
//
//  Input:   tStep = routing time step (sec)
//  Output:  none
//  Purpose: routes water quality constituents through the drainage
//           network over the current time step.
//
{
    int    i, j;
    double qIn, vAvg;

    // --- find mass flow each link contributes to its downstream node
    for ( i=0; i<Nobjects[LINK]; i++ ) findLinkMassFlow(i);

    // --- find new water quality concentration at each node  
    for (j = 0; j < Nobjects[NODE]; j++)
    {
        // --- get node inflow and average volume
        qIn = Node[j].inflow;
        vAvg = (Node[j].oldVolume + Node[j].newVolume) / 2.0;
        
        // --- save inflow concentrations if treatment applied
        if ( Node[j].treatment )
        {
            if ( qIn < ZERO ) qIn = 0.0;
            treatmnt_setInflow(qIn, Node[j].newQual);
        }
       
        // --- find new quality at the node 
        if ( Node[j].type == STORAGE || Node[j].newVolume > FUDGE )
            findStorageQual(j, tStep);
        else findNodeQual(j);

        // --- apply treatment to new quality values
        if ( Node[j].treatment ) treatmnt_treat(j, qIn, vAvg, tStep);
    }

    // --- find new water quality in each link
    for ( i=0; i<Nobjects[LINK]; i++ ) findLinkQual(i, tStep);
}

//=============================================================================

double qualrout_getCstrQual(double c, double v, double wIn, double qIn, 
                            double kDecay, double tStep)
//
//  Input:   c = pollutant concentration at start of time step (mass/ft3)
//           v = average volume over time step (ft3)
//           wIn = mass inflow rate (mass/sec)
//           qIn = inflow rate (cfs)
//           kDecay = first-order decay coeff. (1/sec)
//           tStep = time step (sec)
//  Output:  returns pollutant concentration at end of time step (mass/ft3)
//  Purpose: finds pollutant concentration within a CSTR volume
{
    double f;
    if ( v < ZERO ) return 0.0;
    f = (qIn/v + kDecay)*tStep;
    if ( f <= ZERO || f > 15.0 ) return c;
    else f = exp(-f);
    c = c*f;
    if ( qIn > ZERO ) c = c + wIn/qIn*(1.0-f);
    return c; 
}

//=============================================================================

void findLinkMassFlow(int i)
//
//  Input:   i = link index
//  Output:  none
//  Purpose: adds constituent mass flow out of link to the total
//           accumulation at the link's downstream node.
//
//  Note:    Node[].newQual[], the accumulator variable, already contains
//           contributions from runoff and other external inflows from
//           calculations made in routing_execute().
{
    int    j, p;
    double qLink;

    // --- find inflow to downstream node
    qLink = Link[i].newFlow;

    // --- identify index of downstream node
    j = Link[i].node2;
    if ( qLink < 0.0 ) j = Link[i].node1;
    qLink = fabs(qLink);

    // --- add mass inflow from link to total at downstream node
    for (p = 0; p < Nobjects[POLLUT]; p++)
    {
        Node[j].newQual[p] += qLink * Link[i].oldQual[p];
    }
}

//=============================================================================

void findNodeQual(int j)
//
//  Input:   j = node index
//  Output:  none
//  Purpose: finds new quality in a node with no storage volume.
//
{
    int    p;
    double qNode;

    // --- if there is flow into node then concen. = mass inflow/node flow
    qNode = Node[j].inflow;
    if ( qNode > ZERO )
    {
        for (p = 0; p < Nobjects[POLLUT]; p++)
        {
            Node[j].newQual[p] /= qNode;
        }
    }

    // --- otherwise concen. is 0
    else for (p = 0; p < Nobjects[POLLUT]; p++) Node[j].newQual[p] = 0.0;
}

//=============================================================================

void findLinkQual(int i, double tStep)
//
//  Input:   i = link index
//           tStep = routing time step (sec)
//  Output:  none
//  Purpose: finds new quality in a link at end of the current time step.
//
{
    int    j,                // upstream node index
           k,                // conduit index
           p;                // pollutant index
    double kDecay,           // pollutant decay constant (1/sec)
           massReacted,      // pollutant mass lost to reaction (mass/sec)
           wIn,              // pollutant mass inflow rate (mass/sec)
           qIn,              // inflow rate (cfs)
           qOut,             // outflow rate (cfs)
           vOld,             // link volume at start of time step (ft3)
           vNew,             // link volume at end of time step (ft3)
           vAvg,             // average link volume (ft3)
           c1,               // current concentration within link (mass/ft3)
           c2;               // new concentration within link (mass/ft3)

    // --- identify index of upstream node
    j = Link[i].node1;
    if ( Link[i].newFlow < 0.0 ) j = Link[i].node2;

    // --- link concentration equals that of upstream node when:
    //     a. link is not a conduit or
    //     b. link is a dummy link or
    //     c. steady flow routing is used
    if ( Link[i].type != CONDUIT || Link[i].xsect.type == DUMMY ||
         RouteModel == SF)
    {
        for (p = 0; p < Nobjects[POLLUT]; p++)
        {
            Link[i].newQual[p] = Node[j].newQual[p];
        }
        return;
    }

    // --- find inlet & outlet flow
    k = Link[i].subIndex;
    qIn  = fabs(Conduit[k].q1) * (double)Conduit[k].barrels;
    qOut = fabs(Conduit[k].q2) * (double)Conduit[k].barrels;

    // --- get old, new and average volumes
    vOld = Link[i].oldVolume;
    vNew = Link[i].newVolume;
    vAvg = (vOld + vNew) / 2.0;

    // --- for DW routing, adjust inflow to compensate for volume change
    if ( RouteModel == DW )
    {
        qIn = qIn + (vNew - vOld) / tStep; 
        qIn = MAX(qIn, 0.0);
    }

    // --- for each pollutant
    for (p = 0; p < Nobjects[POLLUT]; p++)
    {
        // --- assign values to local variables
        c1 = Link[i].oldQual[p];
        kDecay = Pollut[p].kDecay;
        wIn = Node[j].newQual[p]*qIn;

        // --- use CSTR eqn. to find new concentration
        c2 = qualrout_getCstrQual(c1, vAvg, wIn, qIn, kDecay, tStep);

        // --- update overall mass balance if mass reacted        
        if ( kDecay != 0.0 )
        {
            massReacted = (c1+c2)/2.0 * vAvg * kDecay;
            massbal_addReactedMass(p, massReacted);
        }

        // --- assign new concen. to link
        c2 = MAX(c2, 0.0);
        Link[i].newQual[p] = c2;
    }
}

//=============================================================================

void  findStorageQual(int j, double tStep)
//
//  Input:   j = node index
//           tStep = routing time step (sec)
//  Output:  none
//  Purpose: finds new quality in a node with storage volume.
//  
{
    int    p;                // Pollutant index
    double kDecay,           // pollutant decay constant (1/sec)
           massReacted,      // pollutant mass lost to reaction (mass/sec)
           qIn,              // inflow rate (cfs)
           vAvg,             // avg. volume over time step (ft3)
           wIn,              // pollutant mass inflow rate (mass)
           c1,               // initial pollutant concentration (mass/ft3)
           c2;               // final pollutant concentration (mass/ft3)


    // --- get inflow rate & average volume
    qIn = Node[j].inflow;
    vAvg = (Node[j].oldVolume + Node[j].newVolume) / 2.0;

    // --- update hydraulic residence time for storage nodes
    //     (HRT can be used in treatment functions)
    if ( Node[j].type == STORAGE ) updateHRT(j, Node[j].oldVolume, qIn, tStep);

    // --- for each pollutant
    for (p = 0; p < Nobjects[POLLUT]; p++)
    {
        // --- get current concentration 
        c1 = Node[j].oldQual[p];

        // --- mass inflow rate is temporarily saved in Node[j].newQual[p]
        wIn = Node[j].newQual[p];

        // --- apply any 1st order decay only if no separate treatment function
        if ( Node[j].treatment == NULL || 
             Node[j].treatment[p].equation == NULL ) kDecay = Pollut[p].kDecay;
        else kDecay = 0.0;

        // --- use CSTR eqn. to find new concentration
        c2 = qualrout_getCstrQual(c1, vAvg, wIn, qIn, kDecay, tStep);

        // --- update overall mass balance if mass reacted        
        if ( kDecay != 0.0 )
        {
            massReacted = (c1+c2)/2.0 * vAvg * kDecay;
            massbal_addReactedMass(p, massReacted);
        }

        // --- assign new concen. to node
        c2 = MAX(c2, 0.0);
        Node[j].newQual[p] = c2;
    }
}

//=============================================================================

void updateHRT(int j, double v, double q, double tStep)
//
//  Input:   j = node index
//           v = storage volume (ft3)
//           q = inflow rate (cfs)
//           tStep = time step (sec)
//  Output:  none
//  Purpose: updates hydraulic residence time (i.e., water age) at a 
//           storage node.
//
{
    int    k = Node[j].subIndex;
    double hrt = Storage[k].hrt;

    if ( v < ZERO ) hrt = 0.0;
    else hrt = (hrt + tStep) * v / (v + q*tStep);
    Storage[k].hrt = MAX(hrt, 0.0);
}

//=============================================================================
